/*
 * Copyright (c) 2007 Christian W. Guenther (christian@deckfour.org)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * License to link and use is also granted to open source programs which are not
 * licensed under the terms of the GPL, given that they satisfy one or more of
 * the following conditions: 1) Explicit license is granted to the ProM and
 * ProMimport programs for usage, linking, and derivative work. 2) Carte blance
 * license is granted to all programs developed at Eindhoven Technical
 * University, The Netherlands, or under the umbrella of STW Technology
 * Foundation, The Netherlands. For further exemptions not covered by the above
 * conditions, please contact the author of this code.
 */
package org.processmining.models.graphbased.directed.fuzzymodel.transform;

import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.events.Logger.MessageLevel;
import org.processmining.models.graphbased.directed.fuzzymodel.MutableFuzzyGraph;

/**
 * @author christian
 * @author Jiafei Li
 * 
 */
public class ConcurrencyEdgeTransformer extends FuzzyGraphTransformer {

	/*
	 * HV 2018-07-26
	 * The graph to transform. Changes to the graph are to be made in this graph.
	 */
	protected MutableFuzzyGraph graph;
	/*
	 * HV 2018-07-26
	 * A copy of the graph before it was transformed. When checking whether the graph needs to be changed, this graph should be used.
	 */
	private MutableFuzzyGraph initialGraph;
	protected double preserveThreshold;
	protected double ratioThreshold;
	protected int counterParallelized;
	protected int counterResolved;

	public ConcurrencyEdgeTransformer(PluginContext context) {
		super(context, "Concurrency edge transformer");
		graph = null;
		initialGraph = null;
		preserveThreshold = 0.4;
		ratioThreshold = 0.7;
	}

	public void setPreserveThreshold(double threshold) {
		preserveThreshold = threshold;
	}

	public void setRatioThreshold(double threshold) {
		ratioThreshold = threshold;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.mining.fuzzymining.graph.transform.FuzzyGraphTransformer
	 * #transform(org.processmining.mining.fuzzymining.graph.FuzzyGraph)
	 */
	public void transform(MutableFuzzyGraph graph) {
		counterParallelized = 0;
		counterResolved = 0;
		this.graph = graph;
		/*
	     * HV 2018-07-26
		 * Create a copy of the graph before it is transformed.
		 */
		this.initialGraph = (MutableFuzzyGraph) graph.clone();
		int numberOfNodes = graph.getNumberOfInitialNodes();
		for (int x = 0; x < numberOfNodes; x++) {
			for (int y = 0; y < x; y++) {
				processRelationPair(x, y);
			}
		}
		context.log("    Concurrency edge transformer: Resolved " + counterResolved + " conflicts, parallelized "
				+ counterParallelized + " node pairs.", MessageLevel.NORMAL);
	}

	protected void processRelationPair(int indexA, int indexB) {
		/*
		 * HV 2018-07-26
		 * Check whether a transformation is needed on the original (non-transformed) graph.
		 */
		double sigFwd = initialGraph.getBinarySignificance(indexA, indexB);
		double sigRwd = initialGraph.getBinarySignificance(indexB, indexA);
		if ((sigFwd > 0.0) && (sigRwd > 0.0)) {
			// conflict situation
			double relImpAB = getRelativeImportanceForEndNodes(indexA, indexB);
			double relImpBA = getRelativeImportanceForEndNodes(indexB, indexA);
			if ((relImpAB > preserveThreshold) && (relImpBA > preserveThreshold)) {
				// preserve both links, as they are sufficiently locally important
			} else {
				// investigate ratio of local importance between both relations
				double ratio = Math.min(relImpAB, relImpBA) / Math.max(relImpAB, relImpBA);
				if (ratio < ratioThreshold) {
					// preserve locally more important relation
					if (relImpAB > relImpBA) {
						// erase B -> A
						graph.setBinarySignificance(indexB, indexA, 0.0);
						graph.setBinaryCorrelation(indexB, indexA, 0.0);
					} else {
						// erase A -> B
						graph.setBinarySignificance(indexA, indexB, 0.0);
						graph.setBinaryCorrelation(indexA, indexB, 0.0);
					}
					counterResolved++;
				} else {
					// erase both.
					// erase A -> B
					graph.setBinarySignificance(indexA, indexB, 0.0);
					graph.setBinaryCorrelation(indexA, indexB, 0.0);
					// erase B -> A
					graph.setBinarySignificance(indexB, indexA, 0.0);
					graph.setBinaryCorrelation(indexB, indexA, 0.0);
					counterParallelized++;
				}
			}
		}
		// else: no conflict to be resolved
	}

	protected double getRelativeImportanceForEndNodes(int fromIndex, int toIndex) {
		/*
		 * HV 2018-07-26
		 * Use the original (non-transformed) graph.
		 */
		double sigRef = initialGraph.getBinarySignificance(fromIndex, toIndex);
		// accumulate all outgoing significances of source node, and
		// all incoming significances of target node, respectively.
		double sigSourceOutAcc = 0.0;
		double sigTargetInAcc = 0.0;
		for (int i = initialGraph.getNumberOfInitialNodes() - 1; i >= 0; i--) {
			if (i != fromIndex) { // ignore self-loops in calculation
				sigSourceOutAcc += initialGraph.getBinarySignificance(fromIndex, i);
			}
			if (i != toIndex) { // ignore self-loops in calculation
				sigTargetInAcc += initialGraph.getBinarySignificance(i, toIndex);
			}
		}
		// relative importance is the product of the relative significances
		// within the source's outgoing and the target's incoming links
		double relativeImportance = ((sigRef / sigSourceOutAcc) + (sigRef / sigTargetInAcc));
		return relativeImportance;
	}

}
